#include <msp/gl/renderer.h>
#include "headtrackingcamera.h"
#include "stereocombiner.h"
#include "stereoview.h"

using namespace std;

namespace Msp {
namespace VR {

StereoView::StereoView(const StereoCombiner &c, const GL::Camera &a):
	combiner(c),
	base_camera(a),
	head_camera(0)
{
	init();
}

StereoView::StereoView(const StereoCombiner &c, HeadTrackingCamera &a):
	combiner(c),
	base_camera(a),
	head_camera(&a)
{
	init();
}

void StereoView::init()
{
	content = 0;

	unsigned w = combiner.get_target_width();
	unsigned h = combiner.get_target_height();
	left.create_target(w, h);
	right.create_target(w, h);

	set_eye_spacing(0.064);
	set_strabismus(Geometry::Angle<float>::zero());
}

void StereoView::set_content(const GL::Renderable *r)
{
	content = r;
}

void StereoView::set_eye_spacing(float s)
{
	GL::Vector3 offset(s/2, 0, 0);
	set_eye_matrices(GL::Matrix::translation(-offset), GL::Matrix::translation(offset));
}

void StereoView::set_eye_matrices(const GL::Matrix &left_matrix, const GL::Matrix &right_matrix)
{
	left.offset_matrix = left_matrix;
	right.offset_matrix = right_matrix;
}

void StereoView::set_strabismus(const Geometry::Angle<float> &s)
{
	strabismus = s;
}

void StereoView::setup_frame() const
{
	if(head_camera)
		head_camera->update();

	EyeParams params;
	params.fov = combiner.get_field_of_view();
	if(params.fov==Geometry::Angle<float>::zero())
		params.fov = base_camera.get_field_of_view();

	params.aspect = combiner.get_render_aspect();
	params.near_clip = base_camera.get_near_clip();
	params.far_clip = base_camera.get_far_clip();

	float frustum_skew = combiner.get_frustum_skew();
	float halfw = tan(params.fov/2.0f)*params.aspect;
	frustum_skew = tan(Geometry::atan<float>(frustum_skew*halfw)+strabismus)/halfw;

	left.setup_frame(base_camera, frustum_skew, params);
	right.setup_frame(base_camera, -frustum_skew, params);
}

void StereoView::render() const
{
	combiner.prepare();
	setup_frame();
	if(content)
	{
		content->setup_frame();
		left.render(*content);
		right.render(*content);
		content->finish_frame();
	}
	combiner.render(left.target->color, right.target->color);
}


StereoView::RenderTarget::RenderTarget(unsigned width, unsigned height)
{
	color.set_min_filter(GL::LINEAR);
	color.set_wrap(GL::CLAMP_TO_EDGE);
	color.storage(GL::RGB, width, height);
	fbo.attach(GL::COLOR_ATTACHMENT0, color);

	depth.storage(GL::DEPTH_COMPONENT, width, height);
	fbo.attach(GL::DEPTH_ATTACHMENT, depth);
}


StereoView::Eye::Eye():
	target(0)
{ }

void StereoView::Eye::create_target(unsigned w, unsigned h)
{
	delete target;
	target = new RenderTarget(w, h);
}

void StereoView::Eye::setup_frame(const GL::Camera &base_camera, float frustum_skew, const EyeParams &params) const
{
	GL::Matrix matrix = base_camera.get_object_matrix()*offset_matrix;
	camera.set_position(matrix*GL::Vector3());
	camera.set_up_direction((matrix*GL::Vector4(0, 1, 0, 0)).slice<3>(0));
	camera.set_look_direction((matrix*GL::Vector4(0, 0, -1, 0)).slice<3>(0));

	camera.set_field_of_view(params.fov);
	camera.set_aspect(params.aspect);
	camera.set_depth_clip(params.near_clip, params.far_clip);
	camera.set_frustum_axis(frustum_skew, 0);
}

void StereoView::Eye::render(const GL::Renderable &renderable) const
{
	GL::Bind bind_fbo(target->fbo);
	target->fbo.clear(GL::COLOR_BUFFER_BIT|GL::DEPTH_BUFFER_BIT);
	GL::Renderer renderer(&camera);
	renderable.render(renderer);
}

} // namespace VR
} // namespace Msp
